@@ -69,10 +69,11 @@
  * structure is finally freed.
  *
  * Supported standards:
- * - RFC 1058 - RIPv1
- * - RFC 2453 - RIPv2
- * - RFC 2080 - RIPng
- * - RFC 4822 - RIP cryptographic authentication
+ * RFC 1058 - RIPv1
+ * RFC 2453 - RIPv2
+ * RFC 2080 - RIPng
+ * RFC 2091 - Triggered RIP for demand circuits
+ * RFC 4822 - RIP cryptographic authentication
  */
 
 #include <stdlib.h>
@@ -92,15 +93,6 @@ static void rip_trigger_update(struct rip_proto *p);
  *	RIP routes
  */
 
-static void
-rip_init_entry(struct fib_node *fn)
-{
-  // struct rip_entry *en = (void) *fn;
-
-  const uint offset = OFFSETOF(struct rip_entry, routes);
-  memset((byte *)fn + offset, 0, sizeof(struct rip_entry) - offset);
-}
-
 static struct rip_rte *
 rip_add_rte(struct rip_proto *p, struct rip_rte **rp, struct rip_rte *src)
 {
@@ -116,14 +108,14 @@ rip_add_rte(struct rip_proto *p, struct rip_rte **rp, struct rip_rte *src)
 }
 
 static inline void
-rip_remove_rte(struct rip_proto *p, struct rip_rte **rp)
+rip_remove_rte(struct rip_proto *p UNUSED, struct rip_rte **rp)
 {
   struct rip_rte *rt = *rp;
 
   rip_unlock_neighbor(rt->from);
 
   *rp = rt->next;
-  sl_free(p->rte_slab, rt);
+  sl_free(rt);
 }
 
 static inline int rip_same_rte(struct rip_rte *a, struct rip_rte *b)
@@ -132,6 +124,11 @@ static inline int rip_same_rte(struct rip_rte *a, struct rip_rte *b)
 static inline int rip_valid_rte(struct rip_rte *rt)
 { return rt->from->ifa != NULL; }
 
+struct rip_iface_adata {
+  struct adata ad;
+  struct iface *iface;
+};
+
 /**
  * rip_announce_rte - announce route from RIP routing table to the core
  * @p: RIP instance
@@ -152,27 +149,20 @@ rip_announce_rte(struct rip_proto *p, struct rip_entry *en)
   if (rt)
   {
     /* Update */
-    net *n = net_get(p->p.table, en->n.prefix, en->n.pxlen);
-
     rta a0 = {
-      .src = p->p.main_source,
+      .pref = p->p.main_channel->preference,
       .source = RTS_RIP,
       .scope = SCOPE_UNIVERSE,
-      .cast = RTC_UNICAST
+      .dest = RTD_UNICAST,
     };
 
     u8 rt_metric = rt->metric;
     u16 rt_tag = rt->tag;
-    struct rip_rte *rt2 = rt->next;
 
-    /* Find second valid rte */
-    while (rt2 && !rip_valid_rte(rt2))
-      rt2 = rt2->next;
-
-    if (p->ecmp && rt2)
+    if (p->ecmp)
     {
       /* ECMP route */
-      struct mpnh *nhs = NULL;
+      struct nexthop *nhs = NULL;
       int num = 0;
 
       for (rt = en->routes; rt && (num < p->ecmp); rt = rt->next)
@@ -180,54 +170,76 @@ rip_announce_rte(struct rip_proto *p, struct rip_entry *en)
 	if (!rip_valid_rte(rt))
 	    continue;
 
-	struct mpnh *nh = alloca(sizeof(struct mpnh));
+	struct nexthop *nh = allocz(sizeof(struct nexthop));
+
 	nh->gw = rt->next_hop;
-	nh->iface = rt->from->nbr->iface;
+	nh->iface = rt->from->ifa->iface;
 	nh->weight = rt->from->ifa->cf->ecmp_weight;
-	mpnh_insert(&nhs, nh);
+
+	nexthop_insert(&nhs, nh);
 	num++;
 
 	if (rt->tag != rt_tag)
 	  rt_tag = 0;
       }
 
-      a0.dest = RTD_MULTIPATH;
-      a0.nexthops = nhs;
+      a0.nh = *nhs;
     }
     else
     {
       /* Unipath route */
-      a0.dest = RTD_ROUTER;
-      a0.gw = rt->next_hop;
-      a0.iface = rt->from->nbr->iface;
       a0.from = rt->from->nbr->addr;
+      a0.nh.gw = rt->next_hop;
+      a0.nh.iface = rt->from->ifa->iface;
     }
 
+    struct {
+      ea_list l;
+      eattr e[3];
+      struct rip_iface_adata riad;
+    } ea_block = {
+      .l = { .count = 3, },
+      .e = {
+	{
+	  .id = EA_RIP_METRIC,
+	  .type = EAF_TYPE_INT,
+	  .u.data = rt_metric,
+	},
+	{
+	  .id = EA_RIP_TAG,
+	  .type = EAF_TYPE_INT,
+	  .u.data = rt_tag,
+	},
+	{
+	  .id = EA_RIP_FROM,
+	  .type = EAF_TYPE_IFACE,
+	  .u.ptr = &ea_block.riad.ad,
+	}
+      },
+      .riad = {
+	.ad = { .length = sizeof(struct rip_iface_adata) - sizeof(struct adata) },
+	.iface = a0.nh.iface,
+      },
+    };
+
+    a0.eattrs = &ea_block.l;
+
     rta *a = rta_lookup(&a0);
-    rte *e = rte_get_temp(a);
+    rte *e = rte_get_temp(a, p->p.main_source);
 
-    e->u.rip.from = a0.iface;
-    e->u.rip.metric = rt_metric;
-    e->u.rip.tag = rt_tag;
-
-    e->net = n;
-    e->pflags = 0;
-
-    rte_update(&p->p, n, e);
+    rte_update(&p->p, en->n.addr, e);
   }
   else
   {
     /* Withdraw */
-    net *n = net_find(p->p.table, en->n.prefix, en->n.pxlen);
-    rte_update(&p->p, n, NULL);
+    rte_update(&p->p, en->n.addr, NULL);
   }
 }
 
 /**
  * rip_update_rte - enter a route update to RIP routing table
  * @p: RIP instance
- * @prefix: network prefix
- * @pxlen: network prefix length
+ * @addr: network address
  * @new: a &rip_rte representing the new route
  *
  * The function is called by the RIP packet processing code whenever it receives
@@ -237,9 +249,9 @@ rip_announce_rte(struct rip_proto *p, struct rip_entry *en)
  * rip_withdraw_rte() should be called instead of rip_update_rte().
  */
 void
-rip_update_rte(struct rip_proto *p, ip_addr *prefix, int pxlen, struct rip_rte *new)
+rip_update_rte(struct rip_proto *p, net_addr *n, struct rip_rte *new)
 {
-  struct rip_entry *en = fib_get(&p->rtable, prefix, pxlen);
+  struct rip_entry *en = fib_get(&p->rtable, n);
   struct rip_rte *rt, **rp;
   int changed = 0;
 
@@ -279,8 +291,7 @@ rip_update_rte(struct rip_proto *p, ip_addr *prefix, int pxlen, struct rip_rte *
 /**
  * rip_withdraw_rte - enter a route withdraw to RIP routing table
  * @p: RIP instance
- * @prefix: network prefix
- * @pxlen: network prefix length
+ * @addr: network address
  * @from: a &rip_neighbor propagating the withdraw
  *
  * The function is called by the RIP packet processing code whenever it receives
@@ -288,9 +299,9 @@ rip_update_rte(struct rip_proto *p, ip_addr *prefix, int pxlen, struct rip_rte *
  * removed. Eventually, the change is also propagated by rip_announce_rte().
  */
 void
-rip_withdraw_rte(struct rip_proto *p, ip_addr *prefix, int pxlen, struct rip_neighbor *from)
+rip_withdraw_rte(struct rip_proto *p, net_addr *n, struct rip_neighbor *from)
 {
-  struct rip_entry *en = fib_find(&p->rtable, prefix, pxlen);
+  struct rip_entry *en = fib_find(&p->rtable, n);
   struct rip_rte *rt, **rp;
 
   if (!en)
@@ -317,8 +328,8 @@ rip_withdraw_rte(struct rip_proto *p, ip_addr *prefix, int pxlen, struct rip_nei
  * it into our data structures.
  */
 static void
-rip_rt_notify(struct proto *P, struct rtable *table UNUSED, struct network *net, struct rte *new,
-	      struct rte *old UNUSED, struct ea_list *attrs)
+rip_rt_notify(struct proto *P, struct channel *ch UNUSED, struct network *net, struct rte *new,
+	      struct rte *old UNUSED)
 {
   struct rip_proto *p = (struct rip_proto *) P;
   struct rip_entry *en;
@@ -327,20 +338,22 @@ rip_rt_notify(struct proto *P, struct rtable *table UNUSED, struct network *net,
   if (new)
   {
     /* Update */
-    u32 rt_metric = ea_get_int(attrs, EA_RIP_METRIC, 1);
-    u32 rt_tag = ea_get_int(attrs, EA_RIP_TAG, 0);
+    u32 rt_tag = ea_get_int(new->attrs->eattrs, EA_RIP_TAG, 0);
+    u32 rt_metric = ea_get_int(new->attrs->eattrs, EA_RIP_METRIC, 1);
+    const eattr *rie = ea_find(new->attrs->eattrs, EA_RIP_FROM);
+    struct iface *rt_from = rie ? ((struct rip_iface_adata *) rie->u.ptr)->iface : NULL;
 
     if (rt_metric > p->infinity)
     {
-      log(L_WARN "%s: Invalid rip_metric value %u for route %I/%d",
-	  p->p.name, rt_metric, net->n.prefix, net->n.pxlen);
+      log(L_WARN "%s: Invalid rip_metric value %u for route %N",
+	  p->p.name, rt_metric, net->n.addr);
       rt_metric = p->infinity;
     }
 
     if (rt_tag > 0xffff)
     {
-      log(L_WARN "%s: Invalid rip_tag value %u for route %I/%d",
-	  p->p.name, rt_tag, net->n.prefix, net->n.pxlen);
+      log(L_WARN "%s: Invalid rip_tag value %u for route %N",
+	  p->p.name, rt_tag, net->n.addr);
       rt_metric = p->infinity;
       rt_tag = 0;
     }
@@ -352,21 +365,21 @@ rip_rt_notify(struct proto *P, struct rtable *table UNUSED, struct network *net,
      * collection.
      */
 
-    en = fib_get(&p->rtable, &net->n.prefix, net->n.pxlen);
+    en = fib_get(&p->rtable, net->n.addr);
 
     old_metric = en->valid ? en->metric : -1;
 
     en->valid = RIP_ENTRY_VALID;
     en->metric = rt_metric;
     en->tag = rt_tag;
-    en->from = (new->attrs->src->proto == P) ? new->u.rip.from : NULL;
-    en->iface = new->attrs->iface;
-    en->next_hop = new->attrs->gw;
+    en->from = (new->src->proto == P) ? rt_from : NULL;
+    en->iface = new->attrs->nh.iface;
+    en->next_hop = new->attrs->nh.gw;
   }
   else
   {
     /* Withdraw */
-    en = fib_find(&p->rtable, &net->n.prefix, net->n.pxlen);
+    en = fib_find(&p->rtable, net->n.addr);
 
     if (!en || en->valid != RIP_ENTRY_VALID)
       return;
@@ -384,11 +397,25 @@ rip_rt_notify(struct proto *P, struct rtable *table UNUSED, struct network *net,
   /* Activate triggered updates */
   if (en->metric != old_metric)
   {
-    en->changed = now;
+    en->changed = current_time();
     rip_trigger_update(p);
   }
 }
 
+void
+rip_flush_table(struct rip_proto *p, struct rip_neighbor *n)
+{
+  btime expires = current_time() + n->ifa->cf->timeout_time;
+
+  FIB_WALK(&p->rtable, struct rip_entry, en)
+  {
+    for (struct rip_rte *e = en->routes; e; e = e->next)
+      if ((e->from == n) && (e->expires == TIME_INFINITY))
+	e->expires = expires;
+  }
+  FIB_WALK_END;
+}
+
 
 /*
  *	RIP neighbors
@@ -397,7 +424,7 @@ rip_rt_notify(struct proto *P, struct rtable *table UNUSED, struct network *net,
 struct rip_neighbor *
 rip_get_neighbor(struct rip_proto *p, ip_addr *a, struct rip_iface *ifa)
 {
-  neighbor *nbr = neigh_find2(&p->p, a, ifa->iface, 0);
+  neighbor *nbr = neigh_find(&p->p, *a, ifa->iface, 0);
 
   if (!nbr || (nbr->scope == SCOPE_HOST) || !rip_iface_link_up(ifa))
     return NULL;
@@ -423,7 +450,7 @@ rip_remove_neighbor(struct rip_proto *p, struct rip_neighbor *n)
 {
   neighbor *nbr = n->nbr;
 
-  TRACE(D_EVENTS, "Removing neighbor %I on %s", nbr->addr, nbr->iface->name);
+  TRACE(D_EVENTS, "Removing neighbor %I on %s", nbr->addr, nbr->ifreq->name);
 
   rem_node(NODE n);
   n->ifa = NULL;
@@ -505,7 +532,7 @@ rip_update_bfd(struct rip_proto *p, struct rip_neighbor *n)
     ip_addr saddr = rip_is_v2(p) ? n->ifa->sk->saddr : n->nbr->ifa->ip;
     n->bfd_req = bfd_request_session(p->p.pool, n->nbr->addr, saddr,
 				     n->nbr->iface, p->p.vrf,
-				     rip_bfd_notify, n);
+				     rip_bfd_notify, n, NULL);
   }
 
   if (!use_bfd && n->bfd_req)
@@ -527,14 +554,23 @@ rip_iface_start(struct rip_iface *ifa)
 
   TRACE(D_EVENTS, "Starting interface %s", ifa->iface->name);
 
-  ifa->next_regular = now + (random() % ifa->cf->update_time) + 1;
-  ifa->next_triggered = now;	/* Available immediately */
-  ifa->want_triggered = 1;	/* All routes in triggered update */
-  tm_start(ifa->timer, 1);	/* Or 100 ms */
+  if (! ifa->cf->demand_circuit)
+  {
+    ifa->next_regular = current_time() + (random() % ifa->cf->update_time) + 100 MS;
+    tm_set(ifa->timer, ifa->next_regular);
+  }
+  else
+  {
+    ifa->next_regular = TIME_INFINITY;
+  }
+
   ifa->up = 1;
 
-  if (!ifa->cf->passive)
-    rip_send_request(ifa->rip, ifa);
+  if (ifa->cf->passive)
+    return;
+
+  rip_send_request(p, ifa);
+  rip_send_table(p, ifa, ifa->addr, 0);
 }
 
 static void
@@ -547,10 +583,24 @@ rip_iface_stop(struct rip_iface *ifa)
 
   rip_reset_tx_session(p, ifa);
 
+  ifa->next_regular = 0;
+  ifa->next_triggered = 0;
+  ifa->want_triggered = 0;
+
+  if (ifa->tx_pending)
+    ifa->tx_seqnum++;
+
+  ifa->tx_pending = 0;
+  ifa->req_pending = 0;
+
+  if (ifa->cf->demand_circuit && !ifa->cf->passive)
+    rip_send_flush(p, ifa);
+
   WALK_LIST_FIRST(n, ifa->neigh_list)
     rip_remove_neighbor(p, n);
 
   tm_stop(ifa->timer);
+  tm_stop(ifa->rxmt_timer);
   ifa->up = 0;
 }
 
@@ -651,13 +701,20 @@ rip_add_iface(struct rip_proto *p, struct iface *iface, struct rip_iface_config
   else if (ic->mode == RIP_IM_MULTICAST)
     ifa->addr = rip_is_v2(p) ? IP4_RIP_ROUTERS : IP6_RIP_ROUTERS;
   else /* Broadcast */
-    ifa->addr = iface->addr->brd;
+    ifa->addr = iface->addr4->brd;
+  /*
+   * The above is just a workaround for BSD as it can't send broadcasts
+   * to 255.255.255.255. BSD systems need the network broadcast address instead.
+   *
+   * TODO: move this to sysdep code
+   */
 
   init_list(&ifa->neigh_list);
 
   add_tail(&p->iface_list, NODE ifa);
 
-  ifa->timer = tm_new_set(p->p.pool, rip_iface_timer, ifa, 0, 0);
+  ifa->timer = tm_new_init(p->p.pool, rip_iface_timer, ifa, 0, 0);
+  ifa->rxmt_timer = tm_new_init(p->p.pool, rip_rxmt_timeout, ifa, 0, 0);
 
   struct object_lock *lock = olock_new(p->p.pool);
   lock->type = OBJLOCK_UDP;
@@ -696,7 +753,8 @@ rip_reconfigure_iface(struct rip_proto *p, struct rip_iface *ifa, struct rip_ifa
       (new->port != old->port) ||
       (new->tx_tos != old->tx_tos) ||
       (new->tx_priority != old->tx_priority) ||
-      (new->ttl_security != old->ttl_security))
+      (new->ttl_security != old->ttl_security) ||
+      (new->demand_circuit != old->demand_circuit))
     return 0;
 
   TRACE(D_EVENTS, "Reconfiguring interface %s", ifa->iface->name);
@@ -705,8 +763,20 @@ rip_reconfigure_iface(struct rip_proto *p, struct rip_iface *ifa, struct rip_ifa
 
   rip_iface_update_buffers(ifa);
 
-  if (ifa->next_regular > (now + new->update_time))
-    ifa->next_regular = now + (random() % new->update_time) + 1;
+  if ((! ifa->cf->demand_circuit) &&
+      (ifa->next_regular > (current_time() + new->update_time)))
+    ifa->next_regular = current_time() + (random() % new->update_time) + 100 MS;
+
+  if (ifa->up && new->demand_circuit && (new->passive != old->passive))
+  {
+    if (new->passive)
+      rip_send_flush(p, ifa);
+    else
+    {
+      rip_send_request(p, ifa);
+      rip_send_table(p, ifa, ifa->addr, 0);
+    }
+  }
 
   if (new->check_link != old->check_link)
     rip_iface_update_state(ifa);
@@ -727,7 +797,11 @@ rip_reconfigure_ifaces(struct rip_proto *p, struct rip_config *cf)
 
   WALK_LIST(iface, iface_list)
   {
-    if (! (iface->flags & IF_UP))
+    if (!(iface->flags & IF_UP))
+      continue;
+
+    /* Ignore ifaces without appropriate address */
+    if (rip_is_v2(p) ? !iface->addr4 : !iface->llv6)
       continue;
 
     struct rip_iface *ifa = rip_find_iface(p, iface);
@@ -757,35 +831,34 @@ rip_if_notify(struct proto *P, unsigned flags, struct iface *iface)
 {
   struct rip_proto *p = (void *) P;
   struct rip_config *cf = (void *) P->cf;
+  struct rip_iface *ifa = rip_find_iface(p, iface);
 
   if (iface->flags & IF_IGNORE)
     return;
 
-  if (flags & IF_CHANGE_UP)
+  /* Add, remove or restart interface */
+  if (flags & (IF_CHANGE_UPDOWN | (rip_is_v2(p) ? IF_CHANGE_ADDR4 : IF_CHANGE_LLV6)))
   {
-    struct rip_iface_config *ic = (void *) iface_patt_find(&cf->patt_list, iface, NULL);
+    if (ifa)
+      rip_remove_iface(p, ifa);
 
-    /* For RIPng, ignore ifaces without link-local address */
-    if (rip_is_ng(p) && !ifa_llv6(iface))
+    if (!(iface->flags & IF_UP))
       return;
 
+    /* Ignore ifaces without appropriate address */
+    if (rip_is_v2(p) ? !iface->addr4 : !iface->llv6)
+      return;
+
+    struct rip_iface_config *ic = (void *) iface_patt_find(&cf->patt_list, iface, NULL);
     if (ic)
       rip_add_iface(p, iface, ic);
 
     return;
   }
 
-  struct rip_iface *ifa = rip_find_iface(p, iface);
-
   if (!ifa)
     return;
 
-  if (flags & IF_CHANGE_DOWN)
-  {
-    rip_remove_iface(p, ifa);
-    return;
-  }
-
   if (flags & IF_CHANGE_MTU)
     rip_iface_update_buffers(ifa);
 
@@ -827,24 +900,24 @@ rip_timer(timer *t)
   struct rip_iface *ifa;
   struct rip_neighbor *n, *nn;
   struct fib_iterator fit;
-  bird_clock_t next = now + MIN(cf->min_timeout_time, cf->max_garbage_time);
-  bird_clock_t expires = 0;
+  btime now_ = current_time();
+  btime next = now_ + MIN(cf->min_timeout_time, cf->max_garbage_time);
+  btime expires = 0;
 
   TRACE(D_EVENTS, "Main timer fired");
 
   FIB_ITERATE_INIT(&fit, &p->rtable);
 
   loop:
-  FIB_ITERATE_START(&p->rtable, &fit, node)
+  FIB_ITERATE_START(&p->rtable, &fit, struct rip_entry, en)
   {
-    struct rip_entry *en = (struct rip_entry *) node;
     struct rip_rte *rt, **rp;
     int changed = 0;
 
     /* Checking received routes for timeout and for dead neighbors */
     for (rp = &en->routes; rt = *rp; /* rp = &rt->next */)
     {
-      if (!rip_valid_rte(rt) || (rt->expires <= now))
+      if (!rip_valid_rte(rt) || (rt->expires <= now_))
       {
 	rip_remove_rte(p, rp);
 	changed = 1;
@@ -864,7 +937,7 @@ rip_timer(timer *t)
        * rip_rt_notify() -> p->rtable change, invalidating hidden variables.
        */
 
-      FIB_ITERATE_PUT_NEXT(&fit, &p->rtable, node);
+      FIB_ITERATE_PUT_NEXT(&fit, &p->rtable);
       rip_announce_rte(p, en);
       goto loop;
     }
@@ -874,9 +947,9 @@ rip_timer(timer *t)
     {
       expires = en->changed + cf->max_garbage_time;
 
-      if (expires <= now)
+      if (expires <= now_)
       {
-	// TRACE(D_EVENTS, "entry is too old: %I/%d", en->n.prefix, en->n.pxlen);
+	// TRACE(D_EVENTS, "entry is too old: %N", en->n.addr);
 	en->valid = 0;
       }
       else
@@ -886,36 +959,42 @@ rip_timer(timer *t)
     /* Remove empty nodes */
     if (!en->valid && !en->routes)
     {
-      FIB_ITERATE_PUT(&fit, node);
-      fib_delete(&p->rtable, node);
+      FIB_ITERATE_PUT(&fit);
+      fib_delete(&p->rtable, en);
       goto loop;
     }
   }
-  FIB_ITERATE_END(node);
+  FIB_ITERATE_END;
 
   p->rt_reload = 0;
 
   /* Handling neighbor expiration */
   WALK_LIST(ifa, p->iface_list)
+  {
+    /* No expiration for demand circuit ifaces */
+    if (ifa->cf->demand_circuit)
+      continue;
+
     WALK_LIST_DELSAFE(n, nn, ifa->neigh_list)
       if (n->last_seen)
       {
 	expires = n->last_seen + n->ifa->cf->timeout_time;
 
-	if (expires <= now)
+	if (expires <= now_)
 	  rip_remove_neighbor(p, n);
 	else
 	  next = MIN(next, expires);
       }
+  }
 
-  tm_start(p->timer, MAX(next - now, 1));
+  tm_start(p->timer, MAX(next - now_, 100 MS));
 }
 
 static inline void
 rip_kick_timer(struct rip_proto *p)
 {
-  if (p->timer->expires > (now + 1))
-    tm_start(p->timer, 1);	/* Or 100 ms */
+  if ((p->timer->expires > (current_time() + 100 MS)))
+    tm_start(p->timer, 100 MS);
 }
 
 /**
@@ -933,7 +1012,8 @@ rip_iface_timer(timer *t)
 {
   struct rip_iface *ifa = t->data;
   struct rip_proto *p = ifa->rip;
-  bird_clock_t period = ifa->cf->update_time;
+  btime now_ = current_time();
+  btime period = ifa->cf->update_time;
 
   if (ifa->cf->passive)
     return;
@@ -942,40 +1022,41 @@ rip_iface_timer(timer *t)
 
   if (ifa->tx_active)
   {
-    if (now < (ifa->next_regular + period))
-      { tm_start(ifa->timer, 1); return; }
-
-    /* We are too late, reset is done by rip_send_table() */
-    log(L_WARN "%s: Too slow update on %s, resetting", p->p.name, ifa->iface->name);
+    tm_start(ifa->timer, 100 MS);
+    return;
   }
 
-  if (now >= ifa->next_regular)
+  if (now_ >= ifa->next_regular)
   {
     /* Send regular update, set timer for next period (or following one if necessay) */
     TRACE(D_EVENTS, "Sending regular updates for %s", ifa->iface->name);
     rip_send_table(p, ifa, ifa->addr, 0);
-    ifa->next_regular += period * (1 + ((now - ifa->next_regular) / period));
+    ifa->next_regular += period * (1 + ((now_ - ifa->next_regular) / period));
     ifa->want_triggered = 0;
     p->triggered = 0;
   }
-  else if (ifa->want_triggered && (now >= ifa->next_triggered))
+  else if (ifa->want_triggered && (now_ >= ifa->next_triggered))
   {
     /* Send triggered update, enforce interval between triggered updates */
     TRACE(D_EVENTS, "Sending triggered updates for %s", ifa->iface->name);
     rip_send_table(p, ifa, ifa->addr, ifa->want_triggered);
-    ifa->next_triggered = now + MIN(5, period / 2 + 1);
+    ifa->next_triggered = now_ + MIN(5 S, period / 2);
     ifa->want_triggered = 0;
     p->triggered = 0;
   }
 
-  tm_start(ifa->timer, ifa->want_triggered ? 1 : (ifa->next_regular - now));
+  if (ifa->want_triggered && (ifa->next_triggered < ifa->next_regular))
+    tm_set(ifa->timer, ifa->next_triggered);
+  else if (ifa->next_regular != TIME_INFINITY)
+    tm_set(ifa->timer, ifa->next_regular);
 }
 
+
 static inline void
 rip_iface_kick_timer(struct rip_iface *ifa)
 {
-  if (ifa->timer->expires > (now + 1))
-    tm_start(ifa->timer, 1);	/* Or 100 ms */
+  if ((! tm_active(ifa->timer)) || (ifa->timer->expires > (current_time() + 100 MS)))
+    tm_start(ifa->timer, 100 MS);
 }
 
 static void
@@ -996,11 +1077,10 @@ rip_trigger_update(struct rip_proto *p)
       continue;
 
     TRACE(D_EVENTS, "Scheduling triggered updates for %s", ifa->iface->name);
-    ifa->want_triggered = now;
+    ifa->want_triggered = current_time();
     rip_iface_kick_timer(ifa);
+    p->triggered = 1;
   }
-
-  p->triggered = 1;
 }
 
 
@@ -1008,96 +1088,54 @@ rip_trigger_update(struct rip_proto *p)
  *	RIP protocol glue
  */
 
-static struct ea_list *
-rip_prepare_attrs(struct linpool *pool, ea_list *next, u8 metric, u16 tag)
+static void
+rip_reload_routes(struct channel *C)
 {
-  struct ea_list *l = lp_alloc(pool, sizeof(struct ea_list) + 2 * sizeof(eattr));
-
-  l->next = next;
-  l->flags = EALF_SORTED;
-  l->count = 2;
-
-  l->attrs[0].id = EA_RIP_METRIC;
-  l->attrs[0].flags = 0;
-  l->attrs[0].type = EAF_TYPE_INT | EAF_TEMP;
-  l->attrs[0].u.data = metric;
-
-  l->attrs[1].id = EA_RIP_TAG;
-  l->attrs[1].flags = 0;
-  l->attrs[1].type = EAF_TYPE_INT | EAF_TEMP;
-  l->attrs[1].u.data = tag;
-
-  return l;
-}
-
-static int
-rip_import_control(struct proto *P UNUSED, struct rte **rt, struct ea_list **attrs, struct linpool *pool)
-{
-  /* Prepare attributes with initial values */
-  if ((*rt)->attrs->source != RTS_RIP)
-    *attrs = rip_prepare_attrs(pool, *attrs, 1, 0);
-
-  return 0;
-}
-
-static int
-rip_reload_routes(struct proto *P)
-{
-  struct rip_proto *p = (struct rip_proto *) P;
+  struct rip_proto *p = (struct rip_proto *) C->proto;
 
   if (p->rt_reload)
-    return 1;
+    return;
 
   TRACE(D_EVENTS, "Scheduling route reload");
   p->rt_reload = 1;
   rip_kick_timer(p);
-
-  return 1;
 }
 
-static struct ea_list *
-rip_make_tmp_attrs(struct rte *rt, struct linpool *pool)
+static u32
+rip_rte_igp_metric(struct rte *rt)
 {
-  return rip_prepare_attrs(pool, NULL, rt->u.rip.metric, rt->u.rip.tag);
-}
-
-static void
-rip_store_tmp_attrs(struct rte *rt, struct ea_list *attrs)
-{
-  rt->u.rip.metric = ea_get_int(attrs, EA_RIP_METRIC, 1);
-  rt->u.rip.tag = ea_get_int(attrs, EA_RIP_TAG, 0);
+  return ea_get_int(rt->attrs->eattrs, EA_RIP_METRIC, IGP_METRIC_UNKNOWN);
 }
 
 static int
 rip_rte_better(struct rte *new, struct rte *old)
 {
-  return new->u.rip.metric < old->u.rip.metric;
+  return rip_rte_igp_metric(new) < rip_rte_igp_metric(old);
 }
 
-static int
-rip_rte_same(struct rte *new, struct rte *old)
+static void
+rip_postconfig(struct proto_config *CF)
 {
-  return ((new->u.rip.metric == old->u.rip.metric) &&
-	  (new->u.rip.tag == old->u.rip.tag) &&
-	  (new->u.rip.from == old->u.rip.from));
-}
+  // struct rip_config *cf = (void *) CF;
 
+  /* Define default channel */
+  if (! proto_cf_main_channel(CF))
+    channel_config_new(NULL, net_label[CF->net_type], CF->net_type, CF);
+}
 
 static struct proto *
-rip_init(struct proto_config *cfg)
+rip_init(struct proto_config *CF)
 {
-  struct proto *P = proto_new(cfg, sizeof(struct rip_proto));
+  struct proto *P = proto_new(CF);
+
+  P->main_channel = proto_add_channel(P, proto_cf_main_channel(CF));
 
-  P->accept_ra_types = RA_OPTIMAL;
   P->if_notify = rip_if_notify;
   P->rt_notify = rip_rt_notify;
   P->neigh_notify = rip_neigh_notify;
-  P->import_control = rip_import_control;
   P->reload_routes = rip_reload_routes;
-  P->make_tmp_attrs = rip_make_tmp_attrs;
-  P->store_tmp_attrs = rip_store_tmp_attrs;
   P->rte_better = rip_rte_better;
-  P->rte_same = rip_rte_same;
+  P->rte_igp_metric = rip_rte_igp_metric;
 
   return P;
 }
@@ -1109,10 +1147,12 @@ rip_start(struct proto *P)
   struct rip_config *cf = (void *) (P->cf);
 
   init_list(&p->iface_list);
-  fib_init(&p->rtable, P->pool, sizeof(struct rip_entry), 0, rip_init_entry);
+  fib_init(&p->rtable, P->pool, cf->rip2 ? NET_IP4 : NET_IP6,
+	   sizeof(struct rip_entry), OFFSETOF(struct rip_entry, n), 0, NULL);
   p->rte_slab = sl_new(P->pool, sizeof(struct rip_rte));
-  p->timer = tm_new_set(P->pool, rip_timer, p, 0, 0);
+  p->timer = tm_new_init(P->pool, rip_timer, p, 0, 0);
 
+  p->rip2 = cf->rip2;
   p->ecmp = cf->ecmp;
   p->infinity = cf->infinity;
   p->triggered = 0;
@@ -1126,18 +1166,38 @@ rip_start(struct proto *P)
 }
 
 static int
-rip_reconfigure(struct proto *P, struct proto_config *c)
+rip_shutdown(struct proto *P)
 {
   struct rip_proto *p = (void *) P;
-  struct rip_config *new = (void *) c;
+
+  TRACE(D_EVENTS, "Shutdown requested");
+
+  struct rip_iface *ifa;
+  WALK_LIST(ifa, p->iface_list)
+    rip_iface_stop(ifa);
+
+  return PS_DOWN;
+}
+
+static int
+rip_reconfigure(struct proto *P, struct proto_config *CF)
+{
+  struct rip_proto *p = (void *) P;
+  struct rip_config *new = (void *) CF;
   // struct rip_config *old = (void *) (P->cf);
 
+  if (new->rip2 != p->rip2)
+    return 0;
+
   if (new->infinity != p->infinity)
     return 0;
 
+  if (!proto_configure_channel(P, &P->main_channel, proto_cf_main_channel(CF)))
+    return 0;
+
   TRACE(D_EVENTS, "Reconfiguring");
 
-  p->p.cf = c;
+  p->p.cf = CF;
   p->ecmp = new->ecmp;
   rip_reconfigure_ifaces(p, new);
 
@@ -1148,16 +1208,20 @@ rip_reconfigure(struct proto *P, struct proto_config *c)
 }
 
 static void
-rip_get_route_info(rte *rte, byte *buf, ea_list *attrs UNUSED)
+rip_get_route_info(rte *rte, byte *buf)
 {
-  buf += bsprintf(buf, " (%d/%d)", rte->pref, rte->u.rip.metric);
+  struct rip_proto *p = (struct rip_proto *) rte->src->proto;
+  u32 rt_metric = ea_get_int(rte->attrs->eattrs, EA_RIP_METRIC, p->infinity);
+  u32 rt_tag = ea_get_int(rte->attrs->eattrs, EA_RIP_TAG, 0);
 
-  if (rte->u.rip.tag)
-    bsprintf(buf, " [%04x]", rte->u.rip.tag);
+  buf += bsprintf(buf, " (%d/%d)", rte->attrs->pref, rt_metric);
+
+  if (rt_tag)
+    bsprintf(buf, " [%04x]", rt_tag);
 }
 
 static int
-rip_get_attr(eattr *a, byte *buf, int buflen UNUSED)
+rip_get_attr(const eattr *a, byte *buf, int buflen UNUSED)
 {
   switch (a->id)
   {
@@ -1175,7 +1239,7 @@ rip_get_attr(eattr *a, byte *buf, int buflen UNUSED)
 }
 
 void
-rip_show_interfaces(struct proto *P, char *iff)
+rip_show_interfaces(struct proto *P, const char *iff)
 {
   struct rip_proto *p = (void *) P;
   struct rip_iface *ifa = NULL;
@@ -1184,12 +1248,11 @@ rip_show_interfaces(struct proto *P, char *iff)
   if (p->p.proto_state != PS_UP)
   {
     cli_msg(-1021, "%s: is not up", p->p.name);
-    cli_msg(0, "");
     return;
   }
 
   cli_msg(-1021, "%s:", p->p.name);
-  cli_msg(-1021, "%-10s %-6s %6s %6s %6s",
+  cli_msg(-1021, "%-10s %-6s %6s %6s %7s",
 	  "Interface", "State", "Metric", "Nbrs", "Timer");
 
   WALK_LIST(ifa, p->iface_list)
@@ -1202,16 +1265,16 @@ rip_show_interfaces(struct proto *P, char *iff)
       if (n->last_seen)
 	nbrs++;
 
-    int timer = MAX(ifa->next_regular - now, 0);
-    cli_msg(-1021, "%-10s %-6s %6u %6u %6u",
+    btime now_ = current_time();
+    btime timer = ((ifa->next_regular < TIME_INFINITY) && (ifa->next_regular > now_)) ?
+      (ifa->next_regular - now_) : 0;
+    cli_msg(-1021, "%-10s %-6s %6u %6u %7t",
 	    ifa->iface->name, (ifa->up ? "Up" : "Down"), ifa->cf->metric, nbrs, timer);
   }
-
-  cli_msg(0, "");
 }
 
 void
-rip_show_neighbors(struct proto *P, char *iff)
+rip_show_neighbors(struct proto *P, const char *iff)
 {
   struct rip_proto *p = (void *) P;
   struct rip_iface *ifa = NULL;
@@ -1220,12 +1283,11 @@ rip_show_neighbors(struct proto *P, char *iff)
   if (p->p.proto_state != PS_UP)
   {
     cli_msg(-1022, "%s: is not up", p->p.name);
-    cli_msg(0, "");
     return;
   }
 
   cli_msg(-1022, "%s:", p->p.name);
-  cli_msg(-1022, "%-25s %-10s %6s %6s %6s",
+  cli_msg(-1022, "%-25s %-10s %6s %6s %7s",
 	  "IP address", "Interface", "Metric", "Routes", "Seen");
 
   WALK_LIST(ifa, p->iface_list)
@@ -1238,13 +1300,11 @@ rip_show_neighbors(struct proto *P, char *iff)
       if (!n->last_seen)
 	continue;
 
-      int timer = now - n->last_seen;
-      cli_msg(-1022, "%-25I %-10s %6u %6u %6u",
+      btime timer = current_time() - n->last_seen;
+      cli_msg(-1022, "%-25I %-10s %6u %6u %7t",
 	      n->nbr->addr, ifa->iface->name, ifa->cf->metric, n->uc, timer);
     }
   }
-
-  cli_msg(0, "");
 }
 
 static void
@@ -1255,12 +1315,15 @@ rip_dump(struct proto *P)
   int i;
 
   i = 0;
-  FIB_WALK(&p->rtable, e)
+  FIB_WALK(&p->rtable, struct rip_entry, en)
   {
-    struct rip_entry *en = (struct rip_entry *) e;
-    debug("RIP: entry #%d: %I/%d via %I dev %s valid %d metric %d age %d s\n",
-	  i++, en->n.prefix, en->n.pxlen, en->next_hop, en->iface->name,
-	  en->valid, en->metric, now - en->changed);
+    debug("RIP: entry #%d: %N via %I dev %s valid %d metric %d age %t\n",
+	  i++, en->n.addr, en->next_hop, en->iface ? en->iface->name : "(null)",
+	  en->valid, en->metric, current_time() - en->changed);
+
+    for (struct rip_rte *e = en->routes; e; e = e->next)
+      debug("RIP:   via %I metric %d expires %t\n",
+	    e->next_hop, e->metric, e->expires - current_time());
   }
   FIB_WALK_END;
 
@@ -1277,13 +1340,23 @@ rip_dump(struct proto *P)
 struct protocol proto_rip = {
   .name =		"RIP",
   .template =		"rip%d",
-  .attr_class =		EAP_RIP,
+  .class =		PROTOCOL_RIP,
   .preference =		DEF_PREF_RIP,
+  .channel_mask =	NB_IP,
+  .proto_size =		sizeof(struct rip_proto),
   .config_size =	sizeof(struct rip_config),
+  .postconfig =		rip_postconfig,
   .init =		rip_init,
   .dump =		rip_dump,
   .start =		rip_start,
+  .shutdown =		rip_shutdown,
   .reconfigure =	rip_reconfigure,
   .get_route_info =	rip_get_route_info,
   .get_attr =		rip_get_attr
 };
+
+void
+rip_build(void)
+{
+  proto_build(&proto_rip);
+}
